from typing import Any

from ..transport import Transport


class MockTransport(Transport):
    def __init__(self):
        self.connected = False
        self.closed = False
        self.sent_patterns = []
        self.emitted_patterns = []

    async def connect(self) -> None:
        self.connected = True

    async def close(self) -> None:
        self.closed = True

    async def send(self, pattern: str, data: dict) -> Any:
        self.sent_patterns.append((pattern, data))
        return {"mocked": True, "pattern": pattern, "data": data}

    async def emit(self, pattern: str, data: dict):
        self.emitted_patterns.append((pattern, data))



import json
import uuid

import nats
from nats.aio.client import Client

from ..config.nats import NATSConfig
from ..transport import Transport
from ..utils.parse_response import parse_response


class NATSTransport(Transport):
    config: NATSConfig
    client: Client

    def __init__(self, config: NATSConfig):
        self.config = config

    async def connect(self) -> None:
        self.client = await nats.connect(self.config.servers)

    async def close(self) -> None:
        if not hasattr(self, "client"):
            return
        if not self.client.is_connected:
            return

        await self.client.close()

    async def send(self, pattern: str, data: dict) -> dict:
        correlation_id = str(uuid.uuid4())
        message = json.dumps(
            {"id": correlation_id, "pattern": pattern, "data": data}
        ).encode("utf-8")

        msg = await self.client.request(
            pattern, message, timeout=self.config.response_timeout
        )
        response_body: dict = json.loads(msg.data)
        return parse_response(response_body)

    async def emit(self, pattern: str, data: dict) -> None:
        await self.client.publish(
            pattern,
            json.dumps(
                {"pattern": pattern, "data": data},
            ).encode("utf-8"),
        )



from .transport import Transport


class Client:
    def __init__(self, transport: Transport):
        self.transport = transport

    async def connect(self) -> None:
        await self.transport.connect()

    async def close(self) -> None:
        await self.transport.close()

    async def send(self, pattern: str, data: dict) -> dict:
        return await self.transport.send(pattern, data)

    async def emit(self, pattern: str, data: dict) -> None:
        return await self.transport.emit(pattern, data)

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()



from abc import ABC, abstractmethod
from typing import Any


class Transport(ABC):
    @abstractmethod
    async def connect(self) -> None:
        """
        Estabilish connection to the underlying transport (e.g. broker, socket)
        """
        pass

    @abstractmethod
    async def close(self) -> None:
        """
        Gracefully close the transport connection.
        """
        pass

    @abstractmethod
    async def send(self, pattern: str, data: dict) -> Any:
        """
        RPC-style request/response.
        Client expects a single reply
        """
        pass

    @abstractmethod
    async def emit(self, pattern: str, data: dict) -> None:
        """
        Fire-and-forget event.
        Client does not expect a reply.
        """
        pass



import asyncio
import json
import uuid

from redis.asyncio import Redis

from ..config.redis import RedisConfig
from ..transport import Transport
from ..utils.parse_response import parse_response


class RedisTransport(Transport):
    config: RedisConfig
    client: Redis

    def __init__(self, config: RedisConfig):
        self.config = config

    async def connect(self) -> None:
        self.client = Redis(host=self.config.host, port=self.config.port)

    async def close(self) -> None:
        if hasattr(self, "client") and self.client:
            await self.client.aclose()

    async def send(self, pattern: str, data: dict) -> dict:
        correlation_id = str(uuid.uuid4())
        reply_channel = f"{pattern}.reply"

        message = json.dumps({"id": correlation_id, "pattern": pattern, "data": data})

        pubsub = self.client.pubsub()
        await pubsub.subscribe(reply_channel)

        future = asyncio.get_event_loop().create_future()

        async def listener():
            try:
                async for msg in pubsub.listen():
                    if msg["type"] != "message":
                        continue

                    response_body = json.loads(msg["data"])

                    if response_body.get("id") != correlation_id:
                        continue

                    future.set_result(parse_response(response_body))
                    break
            except Exception as e:
                future.set_exception(e)
            finally:
                await pubsub.unsubscribe(reply_channel)
                await pubsub.aclose()

        asyncio.create_task(listener())
        await self.client.publish(pattern, message)

        return await future

    async def emit(self, pattern: str, data: dict) -> None:
        message = json.dumps({"pattern": pattern, "data": data})
        await self.client.publish(pattern, message)



import json
import uuid

import aio_pika

from ..config.rabbitmq import RabbitMQConfig
from ..transport import Transport
from ..utils.parse_response import parse_response


class RabbitMQTransport(Transport):
    config: RabbitMQConfig
    connection: aio_pika.abc.AbstractRobustConnection
    channel: aio_pika.abc.AbstractChannel

    def __init__(self, config: RabbitMQConfig):
        self.config = config
        self.channel = None  # type: ignore
        self.connection = None  # type: ignore

    async def connect(self) -> None:
        self.connection = await aio_pika.connect_robust(self.config.url)
        self.channel = await self.connection.channel()

    async def close(self) -> None:
        if self.channel:
            await self.channel.close()
        if self.connection:
            await self.connection.close()

    async def send(self, pattern: str, data: dict) -> dict:
        correlation_id = str(uuid.uuid4())
        reply_queue = await self.channel.declare_queue(exclusive=True)

        body = json.dumps(
            {"id": correlation_id, "pattern": pattern, "data": data}
        ).encode("utf-8")

        await self.channel.default_exchange.publish(
            aio_pika.Message(
                body=body,
                correlation_id=correlation_id,
                reply_to=reply_queue.name,
            ),
            routing_key=self.config.queue,
        )

        async with reply_queue.iterator() as queue_iter:
            async for message in queue_iter:
                async with message.process():
                    if message.correlation_id == correlation_id:
                        response_body = json.loads(message.body.decode("utf-8"))

                        return parse_response(response_body)

    async def emit(self, pattern: str, data: dict) -> None:
        body = json.dumps({"pattern": pattern, "data": data}).encode("utf-8")

        await self.channel.default_exchange.publish(
            aio_pika.Message(body=body),
            routing_key=self.config.queue,
        )



import asyncio
import json
import uuid

from ..config.tcp import TCPConfig
from ..transport import Transport
from ..utils.parse_response import parse_response


class TCPTransport(Transport):
    config: TCPConfig
    reader: asyncio.StreamReader
    writer: asyncio.StreamWriter

    def __init__(self, config: TCPConfig):
        self.config = config
        self.reader = None  # type: ignore
        self.writer = None  # type: ignore

    async def connect(self):
        self.reader, self.writer = await asyncio.open_connection(
            self.config.host,
            self.config.port,
        )

    async def close(self):
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()

    async def send(self, pattern: str, data: dict) -> dict:
        correlation_id = str(uuid.uuid4())
        body = json.dumps(
            {"id": correlation_id, "pattern": pattern, "data": data},
        )
        body_with_length = f"{len(body)}#{body}"

        self.writer.write(body_with_length.encode("utf-8"))
        await self.writer.drain()

        buffer = ""
        while True:
            chunk = (await self.reader.read(4096)).decode()
            if not chunk:
                break

            buffer += chunk
            if "#" in buffer:
                break

        length_str, rest = buffer.split("#", 1)
        expected_length = int(length_str)

        while len(rest) < expected_length:
            rest += (await self.reader.read(4096)).decode()

        response_body = json.loads(rest[:expected_length])
        return parse_response(response_body)

    async def emit(self, pattern: str, data: dict) -> None:
        payload = {"pattern": pattern, "data": data}
        body = json.dumps(payload)
        body_with_length = f"{len(body)}#{body}"

        self.writer.write(body_with_length.encode("utf-8"))
        await self.writer.drain()
